@@ -1,9 +1,9 @@
-require File.expand_path('../../../spec_helper', __FILE__)
-require File.expand_path('../shared/terminate.rb', __FILE__)
+require_relative '../../spec_helper'
+require_relative 'shared/terminate'
 require 'strscan'
 
 describe "StringScanner#clear" do
-  it_behaves_like(:strscan_terminate, :clear)
+  it_behaves_like :strscan_terminate, :clear
 
   it "warns in verbose mode that the method is obsolete" do
     s = StringScanner.new("abc")
